/*-
 * Copyright (c) 2010 HIROSHI OOTA. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
 * and other respective copyright holders give permission to link this program
 * with OpenSSL, and distribute the resulting executable, without including
 * the source code for OpenSSL in the source distribution.
 */

/*
 *  --- DO NOT DELETE BELOW THIS LINE ----
 *  $Archive: mod_netdub.a $
 *  $Libraries: -lpthread -liconv$
 */

#include "conf.h"
#include "privs.h"

#define do_PRIVS_ROOT do {PRIVS_ROOT} while(0)
#define do_PRIVS_RELINQUISH do {PRIVS_RELINQUISH} while(0)
#define do_PRIVS_USER do {PRIVS_USER} while(0)
#define do_PRIVS_SETUP(u,g) do {PRIVS_SETUP((u), (g))} while(0)

#include <pthread.h>
#include <signal.h>
#include <iconv.h>
#include "rdns.h"

#include "rev.h"
#define XSTR(x) #x
#define STR(x) XSTR(x)
#define MOD_NETDUB_VERSION		"mod_netdub/0.0" "." STR(REVISION)


static int target_device = -1;
static char status_string[2048];
size_t status_string_size = 0;
/*
 * Command handlers
 */

MODRET netdub_pre_any(cmd_rec *cmd)
{
    
    if (status_string_size <= 0)
      return PR_DECLINED(cmd);
    
    if (!strcmp(cmd->argv[0], C_USER) ||
	!strcmp(cmd->argv[0], C_PASS) ||
	!strcmp(cmd->argv[0], C_RETR) ||
	!strcmp(cmd->argv[0], C_STOR))
      return PR_HANDLED(cmd);

    if (!strcmp(cmd->argv[0], C_TYPE) ||
	!strcmp(cmd->argv[0], C_PASV) ||
	!strcmp(cmd->argv[0], C_EPSV) ||
	!strcmp(cmd->argv[0], C_PORT) ||
	!strcmp(cmd->argv[0], C_EPRT) ||
	!strcmp(cmd->argv[0], C_QUIT))
      return PR_DECLINED(cmd);

    pr_response_add_err(R_550, "%s: %s", cmd->arg, strerror(EPERM));
    return PR_ERROR(cmd);
}

static struct passwd *passwd_dup(pool *p, struct passwd *pw) {
  struct passwd *npw;

  npw = pcalloc(p, sizeof(struct passwd));

  npw->pw_name = pstrdup(p, pw->pw_name);
  npw->pw_passwd = pstrdup(p, pw->pw_passwd);
  npw->pw_uid = pw->pw_uid;
  npw->pw_gid = pw->pw_gid;
  npw->pw_gecos = pstrdup(p, pw->pw_gecos);
  npw->pw_dir = pstrdup(p, pw->pw_dir);
  npw->pw_shell = pstrdup(p, pw->pw_shell);

  return npw;
}

static const char STATUS_FILENAME[] = "$xml$status.xml";
static const char DUBBINGINFO_FILENAME[] = "$netdubbing$dubbinginfo.xml";
static const char COMMAND_FILENAME[] = "$netdubbing$command.xml";
static const char DATA_FILETEMPL[] = "$netdubbing$dev%d.dat";
static const int VIDEO_HEADSIZE = 0x1800;


/*
 * Prepare the data for send.
 */
static size_t prepare_status_string(config_rec *c)
{
    size_t pos = 0;
    int i;

    pos += strlcpy(status_string + pos, 
		   "<?xml version=\"1.0\" encoding=\"shift_jis\"?>\n"
		   "<status>\n"
		   "<dubbing>ready</dubbing>\n"
		   "<accept_dubbing>ready</accept_dubbing>\n",
		   sizeof(status_string) - pos);
    for (i = 0; i < c->argc; i++) {
	if (! strcmp(c->argv[i], "-"))
	  continue;
	pos += snprintf(status_string + pos,
			sizeof(status_string) - pos,
			"<device%d_remain>167772160</device%d_remain>\n"
			"<device%d_recorded>0</device%d_recorded>\n"
			"<device%d_title_remain>362</device%d_title_remain>\n",
			i, i, i, i, i, i);
    }
    pos += strlcpy(status_string + pos,
		   "</status>\n", sizeof(status_string) - pos);
    return pos;
}

extern module auth_module;
static int setup_env(pool *p, cmd_rec *cmd, config_rec *c)
{
    const char *sess_ttyname;
    char *homedir, *xferlog = NULL;;
    struct passwd *pw =
	pr_auth_getpwnam(p, c->argv[1]);
    int res;

    if (pw == NULL || pw->pw_uid == PR_ROOT_UID)
      return FALSE;

    /* Security: other functions perform pw lookups, thus we need to make
     * a local copy of the user just looked up.
     */
    pw = passwd_dup(p, pw);
    
    session.user = pstrdup(p, pw->pw_name);
    session.group = pstrdup(p, pr_auth_gid2name(p, pw->pw_gid));

    session.gids = NULL;
    session.groups = NULL;

    res = pr_auth_getgroups(p, pw->pw_name, &session.gids, &session.groups);
    if (res < 1) {
	pr_log_debug(DEBUG2, "no supplemental groups found for user '%s'",
		     pw->pw_name);
    }
    
    res = pr_auth_is_valid_shell(main_server->conf, pw->pw_shell);
    if (res == FALSE)
      return FALSE;
    
    res = pr_auth_banned_by_ftpusers(main_server->conf, pw->pw_name);
    if (res == TRUE)
      return FALSE;

    /* Attempt to resolve any possible symlinks. */
    do_PRIVS_USER;
    homedir = dir_realpath(p, c->argv[2]);
    do_PRIVS_RELINQUISH;

    if (homedir)
      sstrncpy(session.cwd, homedir, sizeof(session.cwd));
    else
      sstrncpy(session.cwd, c->argv[2], sizeof(session.cwd));


    /* Perform a directory fixup. */
    resolve_deferred_dirs(main_server);
    fixup_dirs(main_server, CF_DEFER);

    sess_ttyname = pr_session_get_ttyname(p);

    do_PRIVS_ROOT;

    log_wtmp(sess_ttyname, session.user, session.c->remote_name,
      session.c->remote_addr);
    session.wtmp_log = TRUE;
    
    if (!xferlog) {
	if (!xferlog)
	  xferlog = get_param_ptr(main_server->conf, "TransferLog", FALSE);
	
	if (!xferlog)
	  xferlog = PR_XFERLOG_PATH;
    }
     
    if (strcasecmp(xferlog, "NONE") == 0) {
	xferlog_open(NULL);
    } else {
	xferlog_open(xferlog);
    }

    res = set_groups(p, pw->pw_gid, session.gids);
    if (res < 0) {
	pr_log_pri(PR_LOG_ERR, "error: unable to set groups: %s",
		   strerror(errno));
    }

    do_PRIVS_RELINQUISH;
    
    /* XXX implements default_root */
	

    do_PRIVS_ROOT;
    
    do_PRIVS_SETUP(pw->pw_uid, pw->pw_gid);

#ifdef HAVE_GETEUID
    if (getegid() != pw->pw_gid ||
	geteuid() != pw->pw_uid) {
	
	PRIVS_RELINQUISH
	    pr_log_pri(PR_LOG_ERR, "error: %s setregid() or setreuid(): %s",
		       session.user, strerror(errno));
	pr_response_send(R_530, _("Login incorrect."));
	
	end_login(1);
    }
#endif

    if (strcmp(session.cwd, "") == 0) {
	pr_log_pri(PR_LOG_ERR, "error: user %s home directory is NULL or \"\"",
		   session.user);
	pr_response_send(R_530, _("Login incorrect."));
	end_login(1);
    }

    /* Attempt to change to the correct directory -- use session.cwd first.
     * This will contain the DefaultChdir directory, if configured...
     */
    if (pr_fsio_chdir_canon(session.cwd, TRUE) == -1) {
	/* Unable to switch to user's real home directory, which is not
	 * allowed.
	 */
	pr_log_pri(PR_LOG_ERR, "%s chdir(\"%s\"): %s", session.user, session.cwd,
		   strerror(errno));
	pr_response_send(R_530, _("Login incorrect."));
	end_login(1);
    }

    
    strlcpy(session.cwd, pr_fs_getcwd(), sizeof(session.cwd));
    strlcpy(session.vwd, pr_fs_getvwd(), sizeof(session.vwd));

    /* Make sure directory config pointers are set correctly */
    dir_check_full(p, cmd, G_NONE, session.cwd, NULL);

    session.proc_prefix = pstrdup(session.pool, session.c->remote_name);
    session.sf_flags = 0;

    /* Update the scoreboard entry */
    pr_scoreboard_entry_update(session.pid,
			       PR_SCORE_USER, session.user,
			       PR_SCORE_CWD, session.cwd,
			       NULL);
    
    pr_session_set_idle();
    
    pr_timer_remove(PR_TIMER_LOGIN, &auth_module);
    
    /* These copies are made from the session.pool, instead of the more
     * volatile pool used originally, in order that the copied data maintain
     * its integrity for the lifetime of the session.
     */
    session.user = pstrdup(session.pool, session.user);
    
    if (session.group)
      session.group = pstrdup(session.pool, session.group);
    
    if (session.gids)
      session.gids = copy_array(session.pool, session.gids);
    
    /* session.groups is an array of strings, so we must copy the string data
     * as well as the pointers.
     */
    session.groups = copy_array_str(session.pool, session.groups);
    
    /* Resolve any deferred-resolution paths in the FS layer */
    pr_resolve_fs_map();
    
    return TRUE;
}

MODRET netdub_user(cmd_rec *cmd)
{
    if (status_string_size > 0)
      return PR_ERROR_MSG(cmd, R_503, _("You are already logged in"));
    return PR_DECLINED(cmd);
}

MODRET netdub_pass(cmd_rec *cmd)
{
    config_rec
	*c = find_config(main_server->conf, CONF_PARAM, "NetDubUserInfo", FALSE);

    pr_log_pri(PR_LOG_DEBUG, "auth_pass is called. %s:%s\n", 
	       (char *)pr_table_get(session.notes, "mod_auth.orig-user", NULL),
	       cmd->arg);
    if (status_string_size > 0)
      return PR_ERROR_MSG(cmd, R_503, _("You are already logged in"));

    if (!c || 
	(strcmp(c->argv[0], "*") && 
	 strcmp(c->argv[0], 
		(char *)pr_table_get(session.notes, "mod_auth.orig-user", NULL))))
      return PR_DECLINED(cmd);

    if (!setup_env(cmd->tmp_pool, cmd, c))
      return PR_DECLINED(cmd);

    c = add_config_param_set(&cmd->server->conf, "authenticated", 1, NULL);
    c->argv[0] = pcalloc(c->pool, sizeof(unsigned char));
    *((unsigned char *) c->argv[0]) = TRUE;

    pr_table_remove(session.notes, "mod_auth.anon-passwd", NULL);
    
    c = find_config(main_server->conf, CONF_PARAM, "NetDubDeviceList", FALSE);

    status_string_size = prepare_status_string(c);
    return PR_HANDLED(cmd);
}

MODRET netdub_pre_retr(cmd_rec *cmd) 
{
    char *dir;

    if (status_string_size <= 0)
      return PR_DECLINED(cmd);

    dir = dir_realpath(cmd->tmp_pool,
		       pr_fs_decode_path(cmd->tmp_pool, cmd->arg));

    if (dir) {
	pr_response_add_err(R_550, "%s: %s", cmd->arg, strerror(EACCES));
	return PR_ERROR(cmd);
    }

    return PR_HANDLED(cmd);
}

MODRET netdub_retr(cmd_rec *cmd)
{
    char *dir = NULL;
    config_rec *c;
    size_t bufsz, nbytes_sent;
    int len;

    if (status_string_size <= 0)
      return PR_DECLINED(cmd);
      
    if (session.restart_pos) {
	pr_response_add_err(R_554, _("%s: invalid REST argument"), cmd->arg);
	return PR_ERROR(cmd);
    }

    if (strcmp(cmd->arg, STATUS_FILENAME) ||
	! (c = find_config(main_server->conf, CONF_PARAM, "NetDubDeviceList", FALSE))) {
	pr_response_add_err(R_550, "%s: %s", cmd->arg, strerror(EACCES));
	return PR_ERROR(cmd);
    }
    
    /* Prepare for any potential throttling. */
    pr_throttle_init(cmd);

    dir = pr_table_get(cmd->notes, "mod_xfer.retr-path", NULL);

    
    /* Send the data */
    pr_data_init(cmd->arg, PR_NETIO_IO_WR);
    session.xfer.path = dir;
    session.xfer.file_size = status_string_size;

    if (pr_data_open(cmd->arg, NULL, PR_NETIO_IO_WR, status_string_size) < 0) {
	pr_data_abort(0, TRUE);
	return PR_ERROR(cmd);
    }

    bufsz =
	main_server->tcp_sndbuf_len > 0 ?
	main_server->tcp_sndbuf_len :
	pr_config_get_xfer_bufsz();
    
    pr_scoreboard_entry_update(session.pid,
			       PR_SCORE_XFER_SIZE, session.xfer.file_size,
			       PR_SCORE_XFER_DONE, (off_t) 0,
			       NULL);

    for (nbytes_sent = 0; nbytes_sent < status_string_size; nbytes_sent += len) {
	int len2;
	pr_signals_handle();
	if (XFER_ABORTED)
	  break;
	len2 = status_string_size - nbytes_sent;

	len = pr_netio_write(session.d->outstrm,
			     status_string + nbytes_sent,
			     len2 < bufsz ? len2 : bufsz);

	pr_throttle_pause(session.xfer.total_bytes, FALSE);
    }

    if (XFER_ABORTED) {
	pr_data_abort(0, FALSE);
	return PR_ERROR(cmd);
    }
    
    pr_throttle_pause(session.xfer.total_bytes, TRUE);

    pr_data_close(FALSE); // send response code(226).
    return PR_HANDLED(cmd);
}

MODRET netdub_pre_stor(cmd_rec *cmd) 
{
    if (status_string_size <= 0)
      return PR_DECLINED(cmd);

    if (strcmp(cmd->arg, DUBBINGINFO_FILENAME) == 0) {
	if (target_device < 0)
	  return PR_HANDLED(cmd);
    } else if (strcmp(cmd->arg, COMMAND_FILENAME) == 0) {
	if (target_device >= 0)
	  return PR_HANDLED(cmd);
    } else if (target_device >= 0) {
	char name[sizeof(DATA_FILETEMPL) + 10];
	snprintf(name, sizeof(name), DATA_FILETEMPL, target_device);
	if (strcmp(cmd->arg, name) == 0)
	  return PR_HANDLED(cmd);
    }
    pr_response_add_err(R_550, "%s: %s", cmd->arg, strerror(EACCES));

    return PR_ERROR(cmd);
}

static off_t read_data(char *buf, off_t bufsz)
{
    off_t pos = 0, len;

    do {
	pr_signals_handle();
	len = pr_netio_read(session.d->instrm, buf + pos, bufsz - pos, 1);
	if (len < 0) {
	    return -1;
	}
	pos += len;
    } while(len > 0 && pos < bufsz);
    return pos;
}

MODRET netdub_stor(cmd_rec *cmd) 
{
    char name[sizeof(DATA_FILETEMPL) + 10], *lbuf;
    off_t bufsz, nbytes_stored;
    pr_fh_t *stor_fh = NULL;    

    if (status_string_size <= 0)
      return PR_DECLINED(cmd);
    
    if (target_device >= 0)
      snprintf(name, sizeof(name), DATA_FILETEMPL, target_device);
    else
      name[0] = 0;

    pr_data_init(cmd->arg, PR_NETIO_IO_RD);

    if (pr_data_open(cmd->arg, NULL, PR_NETIO_IO_RD, 0) < 0) {
	pr_data_abort(0, TRUE);
	return PR_ERROR(cmd);
    }
    
    bufsz = (main_server->tcp_rcvbuf_len > 0 ?  main_server->tcp_rcvbuf_len :
	     pr_config_get_xfer_bufsz());
    bufsz = bufsz < VIDEO_HEADSIZE ? VIDEO_HEADSIZE : bufsz;
    
    lbuf = (char *) palloc(cmd->tmp_pool, bufsz);

    if (!strcmp(cmd->arg, DUBBINGINFO_FILENAME)) {
	config_rec *c =
	    find_config(main_server->conf, CONF_PARAM, "NetDubDeviceList", FALSE);

	off_t len;
	char *p;
	int n;

	len = read_data(lbuf, bufsz);
	if (len >= bufsz) {
	    pr_data_abort(0, TRUE);
	    return PR_ERROR(cmd);
	}
	lbuf[len] = 0;

	if (!(p = strstr(lbuf, "<device>")) ||
	    sscanf(p, "<device>%d</device>", &n) != 1 ||
	    c->argc < n || !strcmp(c->argv[n], "-")) {
	    pr_data_abort(0, TRUE);
	    return PR_ERROR(cmd);
	}
	target_device = n;

	pr_data_close(FALSE);
	nbytes_stored = len;

    } else if(!strcmp(cmd->arg, COMMAND_FILENAME)) {
	off_t len;

	len = read_data(lbuf, bufsz);
	if (len >= bufsz) {
	    pr_data_abort(0, TRUE);
	    return PR_ERROR(cmd);
	}
	lbuf[len] = 0;
	
	if (strstr(lbuf, "<power>off</power>")) {
	    config_rec
		*c = find_config(main_server->conf, CONF_PARAM,
				 "NetDubShutdownCommand", FALSE);
	    pr_log_pri(PR_LOG_INFO, "power off request");
pr_log_pri(PR_LOG_ERR, "performing shutdown");
	    if (c) {
		pid_t pid;

		switch(pid = fork()) {
		default:
		    {
			int status;
			char msg[256];
			
			waitpid(pid, &status, 0);
			if (WIFSIGNALED(status))
			  snprintf(msg, sizeof(msg), "%s: signaled(%d). %s", 
				   (char *)c->argv[0], WTERMSIG(status),
				   WCOREDUMP(status) ? "core dumped." : "");
			else if (WEXITSTATUS(status) != 0)
			  snprintf(msg, sizeof(msg), "%s: error exit(%d).", 
				   (char *)c->argv[0], WEXITSTATUS(status));
			else
			  msg[0] = 0;
			if (msg[0])
			  pr_log_pri(PR_LOG_ERR, msg);
		    }
		    break;
		case 0:
		    {
			int i;
			char *argv[21];
			for (i = 0; i < c->argc; i++) {
 pr_log_pri(PR_LOG_ERR, "argv[%d] %s", i, argv[i]);
			    argv[i] = c->argv[i];
			}
			argv[i] = NULL;
			execv((char *)c->argv[0], argv);
			pr_log_pri(PR_LOG_ERR, "%s: %s", (char *)c->argv[0], 
				   strerror(errno));
			exit(1);
		    }
		    break;
		case -1:
		    pr_log_pri(PR_LOG_ERR, "fork: %s", strerror(errno));
		    break;
		}
	    }
	} else
	  pr_log_pri(PR_LOG_ERR, "unknown command %s", lbuf);

	pr_data_close(FALSE);
	nbytes_stored = len;
    } else if (!strcmp(cmd->arg, name)) {
	off_t len, l;
	char nm[65], ofile[128], *dev;
	config_rec *c =
	    find_config(main_server->conf, CONF_PARAM, "NetDubCpInfo", FALSE);
	iconv_t cd;
	int i, err;
	
	len = read_data(lbuf, bufsz);
	memcpy(nm, lbuf + 0x84, 64);
	ofile[0] = nm[65] = 0;
	
	if ((cd = iconv_open(c ? c->argv[1] : "UTF-8", 
			     c ? c->argv[0] : "CP932")) != (iconv_t)(-1)) {
	    size_t il = strlen(nm), ol = sizeof(ofile);
	    const char *ip = nm; char *op = ofile;

	    if (iconv(cd, &ip, &il, &op, &ol) != (size_t)(-1)) {
		char *f, *t, *p;
		config_rec *c =
		    find_config(main_server->conf, CONF_PARAM, 
				"NetDubCharMap", FALSE);
		if (c) {
		    f = c->argv[0];
		    t = c->argv[1];
		} else {
		    f = "/\\:" ;
		    t = ".--";
		}

		for (i = 0; i < ol; i++)
		  if ((p = strchr(f, ofile[i])) != NULL)
		    ofile[i] = t[p - f];

	    }
	    iconv_close(cd);
	}

	if (ofile[0] == 0) {
	    time_t t;
	    struct tm tm;
	    time(&t);
	    localtime_r(&t, &tm);
	    snprintf(ofile, sizeof(ofile), 
		     "%04d%02d%02d", tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday);
	}

	c = find_config(main_server->conf, CONF_PARAM, "NetDubDeviceList", FALSE);

	dev = c->argv[target_device];

    pr_log_pri(PR_LOG_DEBUG, "STORE:%d %d", __LINE__, target_device);

	if (dev[0] == '+' ){
	    /* store with converted filename */

	    for (l = strlen(ofile), i = 0, 
		 strlcpy(ofile + l, ".mpg", sizeof(ofile) - l);
		 (stor_fh = pr_fsio_open(ofile, 
					 O_WRONLY | O_CREAT | O_EXCL)) == NULL;
		 snprintf(ofile + l, sizeof(ofile) - l, " - %05d.mpg", i++))
	      if (errno != EEXIST) {
		  pr_data_abort(0, TRUE);
		  return PR_ERROR(cmd);
	      }
	    if (stor_fh == NULL) {
		pr_data_abort(0, TRUE);
		return PR_ERROR(cmd);
	    }
	    dev++;
	} else if (dev[0] == 0 || dev[0] == '/') {
	    /* store with original filename */
	    strlcpy(ofile, name, sizeof(ofile));
	    if ((stor_fh = pr_fsio_open(ofile,
					O_WRONLY | O_CREAT)) == NULL) {
		pr_data_abort(0, TRUE);
		return PR_ERROR(cmd);
	    }
	}
	if (dev[0] == '/') { // pipe to command
	    int fds[2];
	    pr_fh_t *err_fh = NULL;    
	    char *p = strrchr(ofile, '.');

	    strlcpy(p, ".log", sizeof(ofile) - (p - ofile));
	    if ((err_fh = pr_fsio_open(ofile, O_WRONLY | O_CREAT)) == NULL) {
		pr_data_abort(0, TRUE);
		return PR_ERROR(cmd);
	    }
	    if (pipe(fds) < 0) {
		pr_fsio_close(err_fh);
		pr_fsio_close(stor_fh);
		pr_data_abort(0, TRUE);
		return PR_ERROR(cmd);
	    }
	    switch(fork()) {
	    case -1:
		close(fds[0]);
		close(fds[1]);
		pr_fsio_close(stor_fh);
		pr_fsio_close(err_fh);
		pr_data_abort(0, TRUE);
		return PR_ERROR(cmd);
	    case 0:
		switch (fork < 0) {
		case -1:
		    {
			char msg[64];
			snprintf(msg, sizeof(msg), "fork: %s", strerror(errno));
			pr_fsio_write(err_fh, msg, strlen(msg));
			pr_fsio_close(stor_fh);
			pr_fsio_close(err_fh);
		    }
		    /* fall through */
		default:
		    exit(1);
		    break;
		case 0:
		    dup2(stor_fh->fh_fd, 1);
		    pr_fsio_close(stor_fh);
		    dup2(err_fh->fh_fd, 2);
		    pr_fsio_close(err_fh);

		    dup2(fds[0], 0);
		    close(fds[0]);
		    close(fds[1]);
		    
		    execl(dev, dev, ofile, (char *)0);
		    {
			char msg[PATH_MAX + 32];
			snprintf(msg, sizeof(msg), 
				 "%s: %s", dev, strerror(errno));
			write(2, msg, strlen(msg));
		    }
		    exit(1);
		    break;
		}
	    default:
		dup2(fds[1], stor_fh->fh_fd);
		close(fds[1]);
		close(fds[0]);
		pr_fsio_close(err_fh);
		break;
	    }
	}

	if ((err = pr_fsio_write(stor_fh, lbuf, len)) != len) {
	    pr_data_abort(err < 0 ? errno : EIO, FALSE);
	    return PR_ERROR(cmd);
	}
	nbytes_stored = len;

	while ((len = read_data(lbuf, bufsz)) > 0) {
	    if (XFER_ABORTED)
	      break;

	    if ((err = pr_fsio_write(stor_fh, lbuf, len)) != len) {
		pr_data_abort(err < 0 ? errno : EIO, FALSE);
		return PR_ERROR(cmd);
	    }
	    nbytes_stored += len;
	    pr_throttle_pause(nbytes_stored, FALSE);
	}
	if (XFER_ABORTED) {
	    pr_data_abort(0, 0);
	    return PR_ERROR(cmd);
	} else if(len < 0) {
	    err = ECONNABORTED;

	    if (session.d != NULL &&
		session.d->instrm != NULL) {
		err = PR_NETIO_ERRNO(session.d->instrm);
	    }
	    pr_data_abort(err, FALSE);
	    return PR_ERROR(cmd);
	} 

	/* If no throttling is configured, this does nothing. */
	pr_throttle_pause(nbytes_stored, TRUE);
	pr_fsio_close(stor_fh);
	pr_data_close(FALSE);
    } else {
	pr_response_add_err(R_550, "%s: %s", cmd->arg, strerror(EACCES));
	return PR_ERROR(cmd);
    }
    return PR_HANDLED(cmd);
}

/* Configuration handlers 
 */

/*
 * NetDubUserInfo rd_name pw_name directory
 */

MODRET set_userinfo(cmd_rec *cmd)
{
    CHECK_ARGS(cmd, 3);
    CHECK_CONF(cmd, CONF_ROOT);
    
    if (ServerType != SERVER_STANDALONE) {
	pr_log_pri(PR_LOG_INFO, 
		   "%s: skipped, server is not standalone.", cmd->argv[0]);
    } else {
	add_config_param_str(cmd->argv[0], 3,
			     cmd->argv[1], cmd->argv[2], cmd->argv[3]);
    
	pr_log_pri(PR_LOG_INFO, "%s: configured", cmd->argv[0]);
    }
    return PR_HANDLED(cmd);
}

/*
 * NetDubNameInfo groupname hostname [reuse_addr]
 */

MODRET set_nameinfo(cmd_rec *cmd)
{
    int bool = 0;
    config_rec *c = NULL;

    CHECK_VARARGS(cmd, 2, 3);

    CHECK_CONF(cmd, CONF_ROOT);
    
    if (ServerType != SERVER_STANDALONE) {
	pr_log_pri(PR_LOG_INFO, 
		   "%s: skipped, server is not standalone.", cmd->argv[0]);
    } else {
	c = add_config_param_str(cmd->argv[0], 3, cmd->argv[1], cmd->argv[2], NULL);

	/* reuse flag */
	if (cmd->argc > 2)
	  bool = get_boolean(cmd, 3);
	
	c->argv[2] = pcalloc(c->pool, sizeof(unsigned char));
	*((unsigned char *) c->argv[2]) = bool;
	
	pr_log_pri(PR_LOG_INFO, "%s: configured %s %s", cmd->argv[0], 
		   (char *)c->argv[0], (char *)c->argv[1]);
    }
    return PR_HANDLED(cmd);
}

/*
 * NetDubDeviceList dev0-attr [dev1-attr [dev2-attr ...]]
 * 
 * dev0 : HDD0 dev1 : DVD dev2 : HDD1
 *
 */
MODRET set_devicelist(cmd_rec *cmd)
{
    config_rec *c;
    int i;

    CHECK_VARARGS(cmd, 1, 10);

    CHECK_CONF(cmd, CONF_ROOT);

    if (ServerType != SERVER_STANDALONE) {
	pr_log_pri(PR_LOG_INFO, 
		   "%s: skipped, server is not standalone.", cmd->argv[0]);
    } else {
	c = add_config_param(cmd->argv[0], cmd->argc - 1,
			     NULL, NULL, NULL, NULL, NULL,
			     NULL, NULL, NULL, NULL, NULL);
	for (i = 0; i < cmd->argc; i++)
	  c->argv[i] = pstrdup(c->pool, cmd->argv[i+1]);
	
	pr_log_pri(PR_LOG_INFO, "%s: configured", cmd->argv[0]);
    }
    return PR_HANDLED(cmd);
}

/*
 * NetDubCpInfo fromcode tocode
 */
MODRET set_codepage(cmd_rec *cmd)
{
    CHECK_ARGS(cmd, 3);
    CHECK_CONF(cmd, CONF_ROOT);
    
    if (ServerType != SERVER_STANDALONE) {
	pr_log_pri(PR_LOG_INFO, 
		   "%s: skipped, server is not standalone.", cmd->argv[0]);
    } else {
	add_config_param_str(cmd->argv[0], 2, cmd->argv[1], cmd->argv[2]);
	pr_log_pri(PR_LOG_INFO, "%s: configured", cmd->argv[0]);
    }
    return PR_HANDLED(cmd);
    
}

/*
 * NetDubShutdownCommand command [args ...]
 */
MODRET set_shutdown(cmd_rec *cmd)
{
    CHECK_VARARGS(cmd, 3, 20);
    CHECK_CONF(cmd, CONF_ROOT);
    
    if (ServerType != SERVER_STANDALONE) {
	pr_log_pri(PR_LOG_INFO, 
		   "%s: skipped, server is not standalone.", cmd->argv[0]);
    } else {
	add_config_param_str(cmd->argv[0], cmd->argc - 1, 
			     cmd->argv[1], cmd->argv[2],
			     cmd->argv[3], cmd->argv[4],
			     cmd->argv[5], cmd->argv[6],
			     cmd->argv[7], cmd->argv[8],
			     cmd->argv[9], cmd->argv[10],
			     cmd->argv[11], cmd->argv[12],
			     cmd->argv[13], cmd->argv[14],
			     cmd->argv[15], cmd->argv[16],
			     cmd->argv[17], cmd->argv[18],
			     cmd->argv[19], cmd->argv[20]);
	pr_log_pri(PR_LOG_INFO, "%s: configured", cmd->argv[0]);
    }
    return PR_HANDLED(cmd);
}

/*
 * NetDubCharMap from to
 */
MODRET set_charmap(cmd_rec *cmd)
{
    CHECK_ARGS(cmd, 2);
    CHECK_CONF(cmd, CONF_ROOT);
    
    if (ServerType != SERVER_STANDALONE) {
	pr_log_pri(PR_LOG_INFO, 
		   "%s: skipped, server is not standalone.", cmd->argv[0]);
    } else if (strlen(cmd->argv[1]) != strlen(cmd->argv[2])) {
      CONF_ERROR(cmd, "invalid arguments");
    } else {
	add_config_param_str(cmd->argv[0], 2, cmd->argv[1], cmd->argv[2]);
	pr_log_pri(PR_LOG_INFO, "%s: configured", cmd->argv[0]);
    }
    return PR_HANDLED(cmd);
    
}


/*
 * Initializations
 */

static void netdub_exit_ev(const void *event_data, void *user_data);
static void netdub_startup_ev(const void *event_data, void *user_data);
static void netdub_restart_ev(const void *event_data, void *user_data);
static void netdub_postparse_ev(const void *event_data, void *user_data);

module netdub_module;

static int netdub_init(void) {
    if (pr_event_register(&netdub_module, "core.startup", netdub_startup_ev, NULL)) {
	pr_log_pri(PR_LOG_ERR, "mod_netdub: pr_event_register: core.startup");
	return -1;
    }

    if (pr_event_register(&netdub_module, "core.restart", netdub_restart_ev, NULL)) {
	pr_log_pri(PR_LOG_ERR, "mod_netdub: pr_event_register: core.restart");
	return -1;
    }

    if (pr_event_register(&netdub_module, "core.exit", netdub_exit_ev, NULL)) {
	pr_log_pri(PR_LOG_ERR, "mod_netdub: pr_event_register: core.exit");
	return -1;
    }

    return 0;
}

static int netdub_sess_init(void)
{
    return 0;
}

/* 
 * Module API tables
 */

static conftable netdub_conftab[] = {
  { "NetDubUserInfo",	set_userinfo,	NULL },
  { "NetDubNameInfo",	set_nameinfo,	NULL },
  { "NetDubDeviceList",	set_devicelist,	NULL },
  { "NetDubCpInfo",		set_codepage,	NULL },
  { "NetDubShutdownCommand",	set_shutdown,	NULL },
  { "NetDubCharMap",		set_charmap,	NULL },
  { NULL }
};

/*
 * The command handler table:
 */

static cmdtable netdub_cmdtab[] = {
  { PRE_CMD,	C_ANY,	G_NONE, netdub_pre_any,		FALSE, FALSE },
  { PRE_CMD,	C_RETR, G_NONE, netdub_pre_retr,	TRUE,	FALSE, CL_READ },
  { PRE_CMD,	C_STOR,	G_WRITE,netdub_pre_stor,	TRUE,	FALSE },
  { CMD,	C_STOR,	G_WRITE,netdub_stor,	TRUE,	FALSE, CL_WRITE },
  { CMD,	C_RETR, G_READ, netdub_retr,	TRUE,	FALSE, CL_READ },
  { CMD,	C_USER,	G_NONE,	netdub_user,	FALSE,	FALSE,	CL_AUTH },
  { CMD,	C_PASS,	G_NONE,	netdub_pass,	FALSE,	FALSE,	CL_AUTH },
  { 0,		NULL }
};

module netdub_module = {

  /* Always NULL */
  NULL, NULL,

  /* Module API version (2.0) */
  0x20,

  /* Module name */
  "netdub",

  /* Module configuration directive handlers */
  netdub_conftab,

  /* Module command handlers */
  netdub_cmdtab,

  /* Module authentication handlers (none in this case) */
  NULL,

  /* Module initialization */
  netdub_init,

  /* Session initialization */
  netdub_sess_init,

  /* Module version */
  MOD_NETDUB_VERSION
};

/*
 * RDNS support routines
 */

static void *rdns_func(void *arg)
{
    sigset_t set;

    sigfillset(&set);
    pthread_sigmask(SIG_SETMASK, &set, NULL);

    return rdns_main(arg);
}


static pthread_t rdns_thread;
static int rdns_sock_fd = -1;

static int start_rdns()
{
    int error;
    config_rec *c;
    static struct rdns_arg arg;
    int i, err;

    if (! find_config(main_server->conf, CONF_PARAM, "NetDubUserInfo", FALSE))
      return FALSE;

    if (! (c = find_config(main_server->conf, CONF_PARAM, "NetDubDeviceList", FALSE)))
      return FALSE;

    for (i = arg.device_list = 0; i < c->argc; i++)
      if (strcmp(c->argv[i], "-"))
	arg.device_list |= 1 << i;

    if (! (c = find_config(main_server->conf, CONF_PARAM, "NetDubNameInfo", FALSE)))
      return FALSE;
    
    arg.group_name = c->argv[0];
    arg.server_name = c->argv[1];
    arg.reuse_addr = *((unsigned char *) c->argv[2]);

    if ((err = rdns_create_socket(&arg))) {
	pr_log_pri(PR_LOG_ERR,
		   "mod_netdub: rdns_create_socket: %s\n", strerror(err));
	return FALSE;
    }

    error = pthread_create(&rdns_thread, NULL, rdns_func, &arg);

    if (error) {
	pr_log_pri(PR_LOG_ERR, "mod_netdub: pthread_create: %s", strerror(error));
	close(arg.sock_fd);
	return FALSE;
    }

    rdns_sock_fd = arg.sock_fd;

    pr_log_pri(PR_LOG_INFO, "%s %s: rdns is ready.",
	       arg.group_name, arg.server_name);
    
    return TRUE;
}

static void netdub_exit_ev(const void *event_data, void *user_data)
{
    if (rdns_sock_fd >= 0) {
	pthread_cancel(rdns_thread);
	pthread_join(rdns_thread, NULL);
	close(rdns_sock_fd);
	rdns_sock_fd = -1;
    }
}

static void netdub_restart_ev(const void *event_data, void *user_data)
{
    netdub_exit_ev(event_data, user_data);
}

static void netdub_postparse_ev(const void *event_data, void *user_data)
{
    if (rdns_sock_fd < 0)
      start_rdns();
}

static void netdub_startup_ev(const void *event_data, void *user_data)
{
    static int initialized = 0;

    if (initialized) {
	pr_log_pri(PR_LOG_ERR, "netdub_startup_ev was called twice");
	return;
    }

    start_rdns();

    if (pr_event_register(&netdub_module, "core.postparse", 
			  netdub_postparse_ev, NULL)) {
	pr_log_pri(PR_LOG_ERR, "mod_netdub: pr_event_register: core.postparse" );
	return;
    }
    initialized = 1;
}
